@@ -18,6 +18,8 @@ import traceback
 
 from collections import defaultdict
 from datetime import timedelta
+from functools import reduce
+from operator import mul
 from threading import Thread
 from typing import Any, cast, Dict, List, Optional, Tuple, TypeVar
 
@@ -26,11 +28,20 @@ import torch
 import torch.distributed as dist
 from torch.distributed._shard.sharded_tensor import ShardedTensor
 from torch.nn.parallel import DistributedDataParallel as DDP
+from torchsnapshot.io_preparer import Chunk
+from torchsnapshot.manifest import ChunkedTensorEntry
+from torchsnapshot.serialization import dtype_to_element_size, string_to_dtype
 
 from .dist_store import get_or_create_store, LinearBarrier
 
 from .flatten import flatten, inflate
-from .io_preparer import ObjectBufferConsumer, prepare_read, prepare_write
+from .io_preparer import (
+    ChunkedTensorIOPreparer,
+    get_storage_path,
+    ObjectBufferConsumer,
+    prepare_read,
+    prepare_write,
+)
 from .io_types import ReadIO, ReadReq, StoragePlugin, WriteIO, WriteReq
 from .manifest import (
     Entry,
@@ -55,8 +66,8 @@ from .version import __version__ as torchsnapshot_version
 logger: logging.Logger = logging.getLogger(__name__)
 
 SNAPSHOT_METADATA_FNAME = ".snapshot_metadata"
-
 T = TypeVar("T")
+_CHUNKING_INSTRUCTION_T = Dict[str, List[Chunk]]
 
 
 class Snapshot:
@@ -65,6 +76,7 @@ class Snapshot:
 
     Basic usage:
     ::
+
         # Define the program state
         app_state = {"model": model, "optimizer": optimizer"}
 
@@ -81,14 +93,18 @@ class Snapshot:
         describing the persisted values and the structure of the original state
         dict.
 
-        Comparing with func:`torch.save` and func:`torch.load`, torchsnapshot:
+        Comparing with :py:func:`torch.save` and :py:func:`torch.load`, torchsnapshot:
 
         - Enables efficient random access of persisted model weights.
+
         - Accelerates persistence by parallelizing writes.
+
             - For replicated values, persistence is parallelized across ranks.
+
         - Enables flexible yet robust elasticity (changing world size on
           restore).
 
+
     Elasticity:
 
         Elasticity is implemented via correctly making persisted values
@@ -99,24 +115,33 @@ class Snapshot:
         the categories in [per-rank, replicated, sharded].
 
         per-rank:
+
             By default, all non-sharded values are treated as per-rank.
+
             On save, the value is only saved by the owning rank.
+
             On load, the value is only made available to the same rank.
 
         replicated:
+
             A user can suggest any non-sharded value as replicated via glob
-                patterns.
+            patterns.
+
             On save, the value is only saved once (can be by any rank).
+
             On load, the value is made available to all ranks, including newly
-                joined ranks.
+            joined ranks.
 
         sharded:
+
             Specific types are always treated as sharded (e.g. ShardedTensor).
+
             On save, all shard-owning ranks save their shards.
+
             On load, all shards are made available to all ranks, including
-                newly joined rank. All ranks can read from all shards for
-                restoring the runtime object from persisted values.
-                (ShardedTensor resharding is powered by torch.dist.checkpoint).
+            newly joined rank. All ranks can read from all shards for
+            restoring the runtime object from persisted values.
+            (ShardedTensor resharding is powered by torch.dist.checkpoint).
 
         If all values within a snapshot are either replicated or sharded, the
         snapshot is automatically reshard-able.
@@ -159,7 +184,7 @@ class Snapshot:
             app_state: The program state to take the snapshot from.
             path: The location to save the snapshot.
             pg: The process group for the processes taking the snapshot.
-                When unspecified:
+            When unspecified:
                     - If distributed is initialized, the global process group will be used.
                     - If distributed is not initialized, single process is assumed.
             replicated: A list of glob patterns for hinting the matching paths
@@ -173,7 +198,10 @@ class Snapshot:
         event_loop = asyncio.new_event_loop()
         pg_wrapper = PGWrapper(pg=pg)
         path, replicated = cls._coalesce_path_and_replicated(
-            path=path, pg_wrapper=pg_wrapper, app_state=app_state, replicated=replicated
+            path=path,
+            pg_wrapper=pg_wrapper,
+            app_state=app_state,
+            replicated=replicated or [],
         )
         storage = url_to_storage_plugin_in_event_loop(
             url_path=path, event_loop=event_loop
@@ -223,7 +251,7 @@ class Snapshot:
             app_state: The program state to take the snapshot from.
             path: The location to save the snapshot.
             pg: The process group for the processes taking the snapshot.
-                When unspecified:
+            When unspecified:
                     - If distributed is initialized, the global process group will be used.
                     - If distributed is not initialized, single process is assumed.
             replicated: A list of glob patterns for hinting the matching paths
@@ -232,15 +260,18 @@ class Snapshot:
 
         Returns:
             A handle with which the newly taken snapshot can be obtained via
-                `.wait()`. Note that waiting on the handle is optional. The
-                snapshot will be committed regardless of whether `.wait()` is
-                invoked.
+            `.wait()`. Note that waiting on the handle is optional. The
+            snapshot will be committed regardless of whether `.wait()` is
+            invoked.
         """
         torch._C._log_api_usage_once("torchsnapshot.Snapshot.async_take")
         event_loop = asyncio.new_event_loop()
         pg_wrapper = PGWrapper(pg=pg)
         path, replicated = cls._coalesce_path_and_replicated(
-            path=path, pg_wrapper=pg_wrapper, app_state=app_state, replicated=replicated
+            path=path,
+            pg_wrapper=pg_wrapper,
+            app_state=app_state,
+            replicated=replicated or [],
         )
         storage = url_to_storage_plugin_in_event_loop(
             url_path=path, event_loop=event_loop
@@ -327,15 +358,44 @@ class Snapshot:
 
         object_entries: Dict[str, Entry] = {}
         write_reqs: List[WriteReq] = []
+
+        chunking_instructions: Dict[str, List[Chunk]] = {}
         for logical_path, obj in flattened.items():
+            # Chunk non-sharded tensors only
+            if isinstance(obj, ShardedTensor) or not isinstance(obj, torch.Tensor):
+                continue
+            chunking_instruction = ChunkedTensorIOPreparer.chunk_tensor(obj)
+            chunking_instructions[logical_path] = chunking_instruction
+
+        chunking_instructions, filtered_logical_paths = cls._partition_logical_paths(
+            replicated_paths, chunking_instructions, flattened, pg_wrapper
+        )
+        replicated_set = set(replicated_paths)
+        for logical_path in chunking_instructions:
+            entry, cwrs = ChunkedTensorIOPreparer.prepare_write(
+                storage_path=get_storage_path(
+                    flattened[logical_path],
+                    logical_path,
+                    pg_wrapper.get_rank(),
+                    logical_path in replicated_set,
+                ),
+                tensor=flattened[logical_path],
+                chunking_instruction=chunking_instructions[logical_path],
+            )
+            entry.replicated = logical_path in replicated_set
+            object_entries[logical_path] = entry
+            write_reqs.extend(cwrs)
+
+        for logical_path in filtered_logical_paths:
             entry, item_write_reqs = prepare_write(
-                obj=obj,
+                obj=flattened[logical_path],
                 logical_path=logical_path,
                 rank=pg_wrapper.get_rank(),
-                replicated=logical_path in replicated_paths,
+                replicated=logical_path in replicated_set,
             )
             object_entries[logical_path] = entry
-            write_reqs += item_write_reqs
+            write_reqs.extend(item_write_reqs)
+
         manifest.update(object_entries)
         manifest = cls._gather_manifest(manifest=manifest, pg=pg_wrapper)
 
@@ -378,7 +438,6 @@ class Snapshot:
         available_entries = get_available_entries(
             manifest=self.metadata.manifest, rank=rank
         )
-
         for key in global_keys:
             self._load_stateful(
                 rank=rank,
@@ -434,7 +493,7 @@ class Snapshot:
 
         A path in snapshot metadata follows the following format:
 
-            RANK/STATEFUL_NAME/STATE_DICT_KEY[/NESTED_CONTAINER_KEY ...]
+            ``RANK/STATEFUL_NAME/STATE_DICT_KEY[/NESTED_CONTAINER_KEY...]``
 
         The rank only matters when the persisted object is "per-rank".
         Arbitrary rank can be used when the persisted object is "replicated" or
@@ -481,7 +540,6 @@ class Snapshot:
         storage = url_to_storage_plugin_in_event_loop(
             url_path=self.path, event_loop=event_loop
         )
-
         read_reqs = prepare_read(
             entry=manifest[unranked_path],
             obj_out=obj_out,
@@ -535,17 +593,16 @@ class Snapshot:
     ) -> List[str]:
         rank = pg.get_rank()
         world_size = pg.get_world_size()
-
         replicated_paths = []
         for path, val in flattened.items():
             if any(fnmatch.fnmatch(path, p) for p in replicated) and not isinstance(
                 val, ShardedTensor
             ):
                 replicated_paths.append(path)
-
         # pyre-ignore
         obj_list: List[List[str]] = [None] * world_size
         pg.all_gather_object(obj_list, replicated_paths)
+
         if rank == 0:
             # A path is only treated as replicated if:
             # (1) The path matches one of the patterns specified in `replicated`
@@ -558,25 +615,11 @@ class Snapshot:
             replicated_paths = list(
                 filter(lambda p: path_count[p] == world_size, replicated_paths)
             )
-
-            paths_partition = Snapshot._partition_replicated_paths(
-                replicated_paths, flattened, world_size
-            )
             replicated_paths_list = [replicated_paths]
         else:
-            paths_partition = None
             replicated_paths_list = [[]]
-
-        local_paths_list = [[]]
-        pg.scatter_object_list(local_paths_list, paths_partition, src=0)
         pg.broadcast_object_list(replicated_paths_list, src=0)
-        local_paths = local_paths_list[0]
         replicated_paths = replicated_paths_list[0]
-
-        for path in replicated_paths:
-            if path not in local_paths:
-                del flattened[path]
-
         return replicated_paths
 
     @classmethod
@@ -682,8 +725,8 @@ path "{logical_path}" which was not available to rank {rank}.
         path: str,
         pg_wrapper: PGWrapper,
         app_state: AppState,
-        replicated: Optional[List[str]],
-    ) -> Tuple[str, Optional[List[str]]]:
+        replicated: List[str],
+    ) -> Tuple[str, List[str]]:
 
         rank = pg_wrapper.get_rank()
 
@@ -699,56 +742,99 @@ path "{logical_path}" which was not available to rank {rank}.
             )
 
         # coalesce replicated
-        if replicated is not None:
-            replicated = cls._infer_replicated(replicated, app_state)
-            # pyre-ignore[9]
-            global_replicated: List[List[str]] = [None] * pg_wrapper.get_world_size()
-            pg_wrapper.all_gather_object(global_replicated, replicated)
-
-            replicated = cls._coalesce_replicated(replicated, global_replicated)
-            if set(global_replicated[rank]) != set(replicated):
-                logger.warning(
-                    f"Rank {rank} specified replicated paths: {set(global_replicated[rank])} "
-                    f"different from replicated paths verified across all ranks: {set(replicated)}"
-                )
+        replicated = cls._infer_replicated(replicated, app_state)
+        # pyre-ignore[9]
+        global_replicated: List[List[str]] = [None] * pg_wrapper.get_world_size()
+        pg_wrapper.all_gather_object(global_replicated, replicated)
 
+        replicated = cls._coalesce_replicated(replicated, global_replicated)
+        if set(global_replicated[rank]) != set(replicated):
+            logger.warning(
+                f"Rank {rank} specified replicated paths: {set(global_replicated[rank])} "
+                f"different from replicated paths verified across all ranks: {set(replicated)}"
+            )
         return obj_list[0], replicated
 
+    @classmethod
+    def _partition_logical_paths(
+        cls,
+        replicated_paths: List[str],
+        chunking_instructions: _CHUNKING_INSTRUCTION_T,
+        flattened: Dict[str, Any],
+        pg_wrapper: PGWrapper,
+    ) -> Tuple[_CHUNKING_INSTRUCTION_T, List[str]]:
+        """
+        Returns:
+            Chunking instruction (for chunkable tensors) and paths to write (for nonchunkable objects).
+        """
+        if pg_wrapper.get_rank() == 0:
+            all_partition_results = cls._partition_replicated_paths(
+                replicated_paths=replicated_paths,
+                chunking_instructions=chunking_instructions,
+                world_size=pg_wrapper.get_world_size(),
+            )
+        else:
+            all_partition_results = None
+        # Scatter partition results among ranks
+        scatter_output_list = [None]
+        pg_wrapper.scatter_object_list(
+            scatter_output_list, all_partition_results, src=0
+        )
+        # pyre-ignore
+        partition_result: Tuple[
+            _CHUNKING_INSTRUCTION_T, List[str]
+        ] = scatter_output_list[0]
+
+        # Add non-replicated chunks and paths to the partition result
+        replicated_set = set(replicated_paths)
+        for path in flattened:
+            if path not in replicated_set:
+                if path in chunking_instructions:
+                    partition_result[0][path] = chunking_instructions[path]
+                else:
+                    partition_result[1].append(path)
+        return partition_result
+
     @staticmethod
     def _partition_replicated_paths(
         replicated_paths: List[str],
-        flattened: Dict[str, Any],
+        chunking_instructions: Dict[str, List[Chunk]],
         world_size: int,
-    ) -> List[List[str]]:
-        """Greedily partitions replicated paths based on the size of the tensor.
-        Inspired by: https://github.com/pytorch/pytorch/blob/1a6b97b8a39062e97562dcea662a375bf89a8fad/torch/distributed/optim/zero_redundancy_optimizer.py#L621-L685
+    ) -> List[Tuple[_CHUNKING_INSTRUCTION_T, List[str]]]:
+        """Partitions a list of replicated paths.
+        Returns a list of size `world_size`; each element of the list is a tuple of
+        chunking instructions for nonchunked tensors and list of paths for nonchunked objs,
+        which were assigned to a rank by a greedy partitioning alg.
         """
-        paths_partition = [[] for _ in range(world_size)]
+        partition_results = [({}, []) for i in range(world_size)]
         sizes = [0] * world_size
-
-        replicated_paths_tensors = []
-        replicated_paths_nontensors = []
-
+        chunked_paths: List[Tuple[str, Chunk, int]] = []
+        nonchunked_paths: List[str] = []
         for path in replicated_paths:
-            if not isinstance(flattened[path], ShardedTensor):
-                if isinstance(flattened[path], torch.Tensor):
-                    replicated_paths_tensors.append(path)
-                else:
-                    replicated_paths_nontensors.append(path)
+            if path in chunking_instructions:
+                for chunk in chunking_instructions[path]:
+                    element_size = dtype_to_element_size(string_to_dtype(chunk.dtype))
+                    chunked_paths.append(
+                        (path, chunk, reduce(mul, chunk.sizes) * element_size)
+                    )
+            else:
+                nonchunked_paths.append(path)
 
-        replicated_paths_tensors_sorted = sorted(
-            replicated_paths_tensors,
-            key=lambda t: flattened[t].numel() * flattened[t].element_size(),
-            reverse=True,
-        )
+        chunked_paths.sort(key=lambda t: t[2], reverse=True)
 
-        for path in replicated_paths_tensors_sorted:
+        # Greedily assign replicated chunks among ranks, based on current sizes of ranks
+        for path, chunk, size in chunked_paths:
             min_rank = np.argmin(sizes)
-            paths_partition[min_rank].append(path)
-            sizes[min_rank] += flattened[path].numel() * flattened[path].element_size()
-        for idx, path in enumerate(replicated_paths_nontensors):
-            paths_partition[idx % world_size].append(path)
-        return paths_partition
+            if path in partition_results[min_rank][0]:
+                partition_results[min_rank][0][path].append(chunk)
+            else:
+                partition_results[min_rank][0][path] = [chunk]
+            sizes[min_rank] += size
+
+        # Round-robin assign rest of replicated paths, which correspond to nonchunked objs
+        for idx, path in enumerate(nonchunked_paths):
+            partition_results[idx % world_size][1].append(path)
+        return partition_results
 
     @staticmethod
     def _infer_replicated(replicated: List[str], app_state: AppState) -> List[str]:
@@ -813,16 +899,28 @@ path "{logical_path}" which was not available to rank {rank}.
         replicated_entries = {}
         for manifest in manifests:
             for path, entry in manifest.items():
-                if is_replicated(entry):
+                if not is_replicated(entry):
+                    continue
+                if path in replicated_entries:
+                    if not isinstance(entry, ChunkedTensorEntry):
+                        raise AssertionError(
+                            "Only one rank should emit the entry for a replicated path "
+                            "unless the entry is ChunkedTensorEntry."
+                        )
+                    # Merge chunks across manifests
+                    replicated_entries[path].chunks.extend(entry.chunks)
+                else:
                     replicated_entries[path] = entry
 
+        for path in replicated_entries:
+            if isinstance(replicated_entries[path], ChunkedTensorEntry):
+                replicated_entries[path].chunks.sort(key=lambda x: x.offsets)
+
         for rank, manifest in enumerate(manifests):
             for path, entry in replicated_entries.items():
-                if path not in manifest:
-                    manifest[path] = entry
+                manifest[path] = entry
             for logical_path, entry in manifest.items():
                 global_manifest[os.path.join(str(rank), logical_path)] = entry
-
         return global_manifest
 
 
