@@ -5,13 +5,14 @@
 # This source code is licensed under the BSD-style license found in the
 # LICENSE file in the root directory of this source tree.
 
-# pyre-ignore-all-errors[2]: allow `Any` in type annotations
+# pyre-ignore-all-errors[2]: Allow `Any` in type annotations
 
+import asyncio
 import unittest
 import uuid
 from contextlib import contextmanager
 
-from typing import Any, Dict, Generator, Union
+from typing import Any, Awaitable, Callable, Dict, Generator, TypeVar, Union
 from unittest import mock
 
 import torch
@@ -22,13 +23,13 @@ from torch.distributed._shard.sharded_tensor import ShardedTensor
 def _tensor_eq(lhs: Union[torch.Tensor, ShardedTensor], rhs: Any) -> bool:
     if type(lhs) != type(rhs):
         return False
-    if isinstance(lhs, torch.Tensor):
-        return torch.allclose(lhs, rhs)
-    elif isinstance(lhs, ShardedTensor):
+    if isinstance(lhs, ShardedTensor):
         for l_shard, r_shard in zip(lhs.local_shards(), rhs.local_shards()):
             if not torch.allclose(l_shard.tensor, r_shard.tensor):
                 return False
         return True
+    elif isinstance(lhs, torch.Tensor):
+        return torch.allclose(lhs, rhs)
     else:
         raise AssertionError("The lhs operand must be a Tensor or ShardedTensor.")
 
@@ -102,3 +103,28 @@ def get_pet_launch_config(nproc: int) -> pet.LaunchConfig:
         max_restarts=0,
         monitor_interval=1,
     )
+
+
+T = TypeVar("T")
+
+
+def async_test(coro: Callable[..., Awaitable[T]]) -> Callable[..., T]:
+    """
+    Decorator for testing asynchronous code.
+    Once we drop support for Python 3.7.x, we can use `unittest.IsolatedAsyncioTestCase` instead.
+
+    Usage:
+        class MyTest(unittest.TestCase):
+            @async_test
+            async def test_x(self):
+                ...
+    """
+
+    def wrapper(*args, **kwargs) -> T:
+        loop = asyncio.new_event_loop()
+        try:
+            return loop.run_until_complete(coro(*args, **kwargs))
+        finally:
+            loop.close()
+
+    return wrapper
